/* Copyright (c) 2006-2015 Polybus Systems Corp., Westford, MA, USA */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "sys_basher.h"

long fileTest(char *base,long num, long num_threads,file_info *finfo,long long limit,result rslt[NUM_DISK_OPS][MAX_DIRS][NUM_DISK_MODES][MAX_FILE_SIZES],Boolean verbose)
{
  long i,j,k,l,total_errs=0,num_dirs,size,num_sizes;
  char string[1024],xstring[1024],lstring[256],mstring[256];
  long errors[4][32][128];
  double wbandw[4][32][128];
  double rbandw[4][32][128];
  double bandw[8],usecs;
  double length[32];
  char path[128][512];
  char *status[8],*pass="PASS",*fail="FAIL";

  if(verbose)printf("\nThread %ld of %ld: Running Disk Tests\n",num,num_threads);

  num_dirs=0;
  while(getDirName(num,num_dirs,finfo,path[num_dirs]))
    {
      num_dirs++;
    }

  num_sizes = 0;
  for(size=12;size<MAX_FILE_SIZES;size+=2)
    {
      length[num_sizes] = 0x00000001LL << size;
      if(length[num_sizes] < limit)
	{
	  fileWrite(base,num,size,0L,finfo,wbandw[0][num_sizes]);
	  num_sizes++;
	}
    }
  num_sizes = 0;
  for(size=12;size<MAX_FILE_SIZES;size+=2)
    {
      if(length[num_sizes] < limit)
	{
	  checkFile(base,num,size,0L,0L,finfo,rbandw[0][num_sizes],errors[0][num_sizes]);
	  num_sizes++;
	}
    }

  total_errs = 0;
  for(i=0;i<num_dirs;i++)
    {
      if(verbose)printf("\nThread %ld of %ld, Directory:%s\n",num,num_threads,finfo->path[i]);
      for(j=0;j<num_sizes;j++)
	{
	  iscale(lstring,"B",length[j]);
	  l=0;
	  bandw[l] = rbandw[0][j][i];
	  bandw[l+1] = wbandw[0][j][i];
	  usecs = length[j]/rbandw[0][j][i];
	  updateBandwidth(rbandw[0][j][i],usecs,length[j],&rslt[0][i][0][j]);
	  usecs = length[j]/wbandw[0][j][i];
	  updateBandwidth(wbandw[0][j][i],usecs,length[j],&rslt[1][i][0][j]);
	  sprintf(rslt[0][i][0][j].name,"Disk Read\t");
	  sprintf(rslt[1][i][0][j].name,"Disk Write\t");
	  rslt[0][i][0][j].size = length[j];
	  rslt[1][i][0][j].size = length[j];
	  rslt[0][i][0][j].runs++;
	  rslt[1][i][0][j].runs++;
	  rslt[0][i][0][j].errors += errors[0][j][i];
	  rslt[1][i][0][j].errors += errors[0][j][i];
	  l+=2;
	}
      if(verbose)printf("\n");

      for(j=0;j<num_sizes;j++)
	{
	  iscale(lstring,"B",length[j]);
	  if(errors[0][j][i] == 0L)
	    {
	      status[0] = pass;
	      printf("PASSED:\t%s/dist_test\n",finfo->path[i]);
	    }
	  else
	    {
	      printf("FAILED:\t%s/disk_test\n",finfo->path[i]);
	      total_errs++;
	      status[0] = fail;
	    }
	}
    }
  return(total_errs == 0L);
}

void fileWrite(char *base,long num,long size,long mode,file_info *finfo,double wbandw[128])
{
  FILE *fp;
  long err,findx=0;
  char string[512],estring[512];
  long long i,length;
  long long *indx,value;
  double start,end,wusecs,len;
  string[0] = 0;
  while(getPathName(num,size,findx,finfo,string))
    {
      fp = fopen(string,"w");
      if(fp == NULL)
	{
	  sprintf(estring,"open %s for write",string);
	  perror(estring);
	}
      else
	{
	  length = 0x00000001LL << size;
	  indx = (long long *)base;
	  value = num << 16;
	  value += size;
	  
	  for(i=0;i<length;i+=8)
	    {
	      *indx++ = value;
	      value += 0x0000000100000000LL;
	    }
	  start = getTime();
	  clearerr(fp);
	  fwrite((void *)base,sizeof(char),length,fp);
	  err = ferror(fp);
	  end = getTime();
	  if(err == -1)
	    {
	      sprintf(estring,"write %s",string);
	      perror(estring);
	    }
	  fclose(fp);
	  sync();
	  wusecs = end - start;
	  len = length;
	  wbandw[findx] = len/wusecs;
	}
      findx++;
    }
}

void checkFile(char *base,long num,long size,long mode,char *mstring,file_info *finfo,double rbandw[128],long errors[128])
{
  FILE *fp;
  long i,length,findx=0,total_errors=0;
  long long value,exp,*indx;;
  long err;
  char string[512],xstring[1024],estring[512],lstring[256];
  double start=0,end=0,rusecs,len;
  string[0] = 0;

  while(getPathName(num,size,findx,finfo,string))
    {
      length = 0x00000001LL << size;
      errors[findx] = 0;
      fp = fopen(string,"r");
      if(fp == NULL)
	{
	  sprintf(estring,"open %s for read",string);
	  perror(estring);
	}
      else
	{
	  clearerr(fp);
	  start = getTime();
	  end = start;
	  fread(base,sizeof(char),length,fp);
	  err = ferror(fp);
	  if(err == -1)
	    {
	      sprintf(estring,"read %s",string);
	      perror(estring);
	    }
	  else
	    {

	      end = getTime();
	      indx = (long long *)base;
	      exp = num << 16;
	      exp += size;
	      for(i=0;i<length;i+=8L)
		{
		  value = *indx++;
		  if(exp != value)
		    {
		      errors[findx] =  errors[findx]+1;
		    }
		  exp += 0x0000000100000000LL;
		}
	    }
	  fclose(fp);
	  unlink(string);
	}

      rusecs = end - start;
      len = length;
      rbandw[findx] = len/rusecs;
      findx++;
    }

}

void createDirs(long num_threads,file_info *finfo)
{
  long num_dirs = 0,i;
  char string[1024];

  num_dirs=0;
  while(getDirName(-1L,num_dirs,finfo,string))
    {
      mkdir(string,0x01FFL);
      num_dirs++;
    }

  for(i=1;i<=num_threads;i++)
    {
      num_dirs=0;
      while(getDirName(i,num_dirs,finfo,string))
	{
	  mkdir(string,0x01FFL);
	  num_dirs++;
	}
    }
  sync();
  sync();
}

void deleteDirs(long num_threads,file_info *finfo)
{
  long num_dirs = 0,i;
  char string[1024];
  for(i=1;i<= num_threads;i++)
    {
      num_dirs=0;
      while(getDirName(i,num_dirs,finfo,string))
	{
	  rmdir(string);
	  num_dirs++;
	}
    }
  num_dirs=0;
  while(getDirName(-1L,num_dirs,finfo,string))
    {
      rmdir(string);
      num_dirs++;
    }
  sync();
  sync();
}
